# -*- coding:utf-8 -*-
# @Author    : g1879
# @date      : 2021-08-09
# @email     : g1879@qq.com
# @File      : memo.py
s = """
=============定位符=============
组成：('列名', '属性名', r'正则表达式')
注：列名是在paths中已定义的列

=============Paths 类属性=============
-------------
# 路径的类型，'css'或'xpath'
paths.type

# 共有的关键元素
paths.rows  # 列表行元素的定位路径，必须
paths.cols  # 行元素中列元素的定位路径，字典格式，必须
paths.next_btn  # 下一页或加载更多按钮元素路径，按页面情况使用，非必须

# 翻页式列表页独有属性
paths.pages_count  # 定位符，总页数所在元素路径，非必须

# 滚动式列表页独有属性
paths.container  # 列表容器，必须

    -------------通过字典创建-------------
    paths_dict = {
        # 路径的类型，只能是'css'或'xpath'，非必须
        'type': 'css',
        
        # 行元素路径，必须 
        'rows': 'xpath 或 css selector',
        
        # 列元素相对于行元素的路径，必须
        'cols': {
            'col1': 'xpath 或 css selector',
            'col2': 'xpath 或 css selector',
            ...
        }
        
        # 翻页式页面总页数获取定位符，格式见上一节，非必须
        'pages_count': 定位符,  
        
        # 下一页（翻页式）或加载更多（滚动式）按钮元素路径，非必须
        'next_btn': 'xpath 或 css selector',
        
        # 行元素所在容器路径，滚动式页面专用，使用滚动式页面时必须
        'container': 'xpath 或 css selector',
    }
    
    paths = Paths(paths_dict=paths_dict)

=============Targets 类=============
targets = Targets(paths)
targets.add_target('标题', '列名', '属性名', r'正则表达式')
targets.start_stop_row = (1,)  # 爬取第2行到最后一行

    -------------通过字典创建-------------
    targets_dict = {
        'start_stop': (1, -1),  # 爬取第2到倒数第2行，规则与切片的一样，非必须
        '目标1': '列1',
        '目标2': ('列2', 'href'),
        '目标3': ('列3', 'src', '(.*)@')
        ...
    }
"""
print(s)



# -*- coding:utf-8 -*-
# @Author    : g1879
# @date      : 2020-11-17
# @email     : g1879@qq.com
# @File      : common.py
from typing import Tuple, Union


def format_loc(loc: Union[str, list, tuple]) -> Tuple[str, str, str]:
    """将传入的定位符规范化                                          \n
    :param loc: 简写的定位符：(xpath或css路径, [属性名, 正则表达式])
    :return: 规范的定位符
    """
    re_str = '(.*)'

    if isinstance(loc, str):
        xpath_str, attr_str = loc, 'text'

    elif isinstance(loc, (list, tuple)):
        length = len(loc)

        if length == 1:
            xpath_str, attr_str = loc[0], 'text'
        elif length == 2:
            xpath_str, attr_str = loc
        elif length == 3:
            xpath_str, attr_str, re_str = loc
        else:
            raise ValueError('定位符位数不正确，应为(xpath或css路径, [属性名, 正则语句])。')

    else:
        raise ValueError('只能传入str、list或tuple。')

    return xpath_str, attr_str, re_str



# -*- coding:utf-8 -*-
# @Author    : g1879
# @date      : 2020-11-17
# @email     : g1879@qq.com
# @File      : targets.py
from typing import Union

from .common import format_loc
from .paths import Paths


class Targets(object):
    """此类用于管理待爬取目标                                         \n
    待爬目标由一个3位的元组表示：(列名, [属性名, 正则表达式])              \n
    如：('作者', 'text', '(.*)')                                   \n
    其中后两位可省略，当省略时默认获取元素text值                          \n
    即传入'作者' 与传入 ('作者', 'text', '(.*)')是等价的               \n
    示例：                                                         \n
        - '作者'                                                   \n
        - '链接', 'href'                                           \n
        - '页数', 'text', '(\\d+)'                                 \n
    """

    def __init__(self,
                 paths: Union[Paths, dict],
                 targets: dict = None):
        """初始化                              \n
        :param paths: Paths对象或路径字典
        :param targets: 要爬取的内容字典
        """
        if isinstance(paths, Paths):
            self._paths = paths

        elif isinstance(paths, dict):
            self._paths = Paths(paths_dict=paths)

        else:
            raise TypeError('paths参数只能是Paths或dict类型。')

        self._targets = {}
        self._start_stop_row = (None, None)

        if targets:
            self.set_targets(targets)

    @property
    def paths(self) -> Paths:
        """返回页面路径管理对象"""
        return self._paths

    @property
    def start_stop_row(self) -> tuple:
        """返回起止行号"""
        return self._start_stop_row

    @start_stop_row.setter
    def start_stop_row(self, start_stop: Union[list, tuple]):
        """设置起止行号"""
        if not isinstance(start_stop, (list, tuple)) or not 0 < len(start_stop) < 3:
            raise ValueError

        self._start_stop_row = (start_stop[0], None) if len(start_stop) == 1 else start_stop

    @property
    def targets(self) -> dict:
        """返回所有目标组成的字典"""
        return self._targets

    @targets.setter
    def targets(self, targets: dict) -> None:
        """批量设置目标                            \n
        :param targets: 列表或元组格式的目标信息
        :return: None
        """
        self.set_targets(targets)

    def set_targets(self, targets: dict) -> None:
        """批量设置目标                            \n
        :param targets: 列表或元组格式的目标信息
        :return: None
        """
        if not isinstance(targets, dict):
            raise TypeError(f'请传入dict，不是{type(targets)}')

        # 有范围设置时
        s_s = targets.get('start_stop', None)
        if not s_s:
            self._start_stop_row = (None, None)
        elif isinstance(s_s, (list, tuple)) and isinstance(s_s[0], (int, type(None))):
            if not 0 < len(s_s) < 3:
                raise ValueError('start_stop元素个数应为1-2个。')

            self._start_stop_row = (s_s[0], None) if len(s_s) == 1 else s_s

        for name in targets:
            if name == 'start_stop':
                continue

            xpath_str, attr_str, re_str = format_loc(targets[name])
            self.add_target(name, xpath_str, attr_str, re_str)

    def add_target(self, name: str, col: str, attr: str = None, re_str: str = None) -> None:
        """添加待爬目标                                     \n
        :param name: 目标标题
        :param col: 已在 paths 中定义的列的名称
        :param attr: 要抓取的属性
        :param re_str: 提取属性内容的正则表达式
        :return: None
        """
        path_str = self.paths.cols[col]
        attr = attr or 'text'
        re_str = re_str or '(.*)'

        self._targets[name] = (path_str, attr, re_str)



# -*- coding:utf-8 -*-
# @Author    : g1879
# @date      : 2020-08-18
# @email     : g1879@qq.com
# @File      : scrolling_page.py
# 依赖：DrissionPage >= v1.1.3
from time import sleep
from typing import Union, List

from DataRecorder import Recorder
from DrissionPage import Drission
from DrissionPage.driver_element import DriverElement

from .base_page import BasePage
from .paths import Paths
from .targets import Targets


class ScrollingPage(BasePage):
    """滚动式列表页基类
    列表页基类提取了列表页面共有的特征，即栏目名、数据行容器、数据行、列，
    封装了对页面的基本读取和操作方法
    """

    def __init__(self,
                 paths: Union[Paths, dict],
                 index_url: str = None,
                 timeout: float = 10,
                 drission: Drission = None):
        """初始化函数                      \n
        :param drission: 驱动器对象
        :param index_url: 首页url
        :param paths: 页面元素定位语句
        """
        super().__init__(paths, index_url, 'd', timeout, drission)
        self.index_url = index_url

        self.container = self.ele(f'{self.paths.type}:{self.paths.container}')
        self._last_row = None

    def get_list(self,
                 targets: Union[Targets, dict],
                 scroll_times: int = 0,
                 wait: float = 0,
                 show_msg: bool = True,
                 recorder: Recorder = None,
                 return_data: bool = True) -> List[dict]:
        """获取若干次滚动后的数据列表            \n
        :param targets: 要爬取的内容
        :param scroll_times: 滚动次数
        :param wait: 滚动后等待几秒
        :param show_msg: 是否显示爬取信息
        :param recorder: 记录器对象
        :param return_data: 是否返回数据
        :return: 结果列表
        """
        if show_msg:
            print('第1页')

        data_list = []
        first_list = self.get_current_list(targets, show_msg)

        if return_data:
            data_list = first_list

        if recorder:
            recorder.add_data(data_list)

        for i in range(scroll_times):
            self.to_next_page(wait)

            if show_msg:
                print(f'\n第{i + 2}页')

            new_list = self.get_new_list(targets, show_msg)

            if not new_list:
                self.click_more_btn(wait=wait)
                new_list = self.get_new_list(targets, show_msg)

            if not new_list:
                break

            if return_data:
                data_list.extend(new_list)

            if recorder:
                recorder.add_data(new_list)

        if recorder:
            recorder.record()

        return data_list

    def get_current_rows(self) -> List[DriverElement]:
        """返回当前页行对象"""
        rows = self.container.eles(f'{self.paths.type}:{self.paths.rows}')

        if rows:
            self._last_row = rows[-1]

        return rows

    def get_new_list(self, targets: Union[Targets, dict], show_msg: bool = True) -> List[dict]:
        """获取新加载的数据列表                      \n
        :param targets: 要爬取的目标对象或字典
        :param show_msg: 是否显示爬取到的内容
        :return: 数据列表
        """
        rows = self.get_new_rows()
        return self._get_list(rows, targets, show_msg)

    def get_new_rows(self) -> List[DriverElement]:
        """获取新加载的行对象"""
        if self._last_row is None:
            return self.get_current_rows()

        if self.paths.type == 'xpath':
            xpath_str = f'xpath:/following::{self.paths.rows.lstrip("./")}'
            rows = self._last_row.eles(xpath_str, timeout=2)
            js = 'return arguments[1].contains(arguments[0])'

            while rows and not rows[-1].run_script(js, self.container.inner_ele):
                rows.pop()

        else:
            rows = self.container.eles(f'css:{self.paths.rows}')
            js = 'return arguments[1]==arguments[0]'

            for key, row in enumerate(rows[::-1]):
                if row.run_script(js, self._last_row.inner_ele):
                    if key != 0:
                        rows = rows[0 - key:]
                    else:
                        rows = []

                    break

        if rows:
            self._last_row = rows[-1]

        return rows

    def click_more_btn(self, wait: float = 1) -> None:
        """点击加载更多按钮（如有）                      \n
        :param wait: 点击按钮后等待秒数
        :return: None
        """
        if self.paths.next_btn:
            next_btn = self.ele(f'{self.paths.type}:{self.paths.next_btn}')

            if next_btn and next_btn.is_displayed():
                next_btn.click(by_js=True)
                sleep(wait)

    def to_next_page(self, wait: float = 1) -> None:
        """滚动一页                      \n
        :param wait: 滚动后等待秒数
        :return: None
        """
        self.scroll_to('bottom')
        sleep(wait)
